# 13.3 球员推荐

现在有这样一个需求，需要从球员库中找到与某位球员相似度比较高的几位球员。那这样一个需求应该怎样实现呢？我们可以回忆一下，本书中提到的算法中，哪个算法是与相似度有关的。没错！就是k近邻算法！所以我们只需要使用k近邻算法来找出离某位球员距离最近的 k 位球员就行了。不过在使用k近邻算法之前，我们需要对数据进行一些处理。

首先是删除一些无用的特征。

```python
# 这些特征并没有太多用处，所以删掉。例如上一节所提到的球衣号码
datafr.drop(['ID','Unnamed: 0','Weak Foot','Release Clause','Wage','Photo', 'Nationality', 'Flag', 'Club Logo', 'International Reputation', 'Body Type', 'Real Face','Jersey Number', 'Joined','LS', 'ST', 'RS', 'LW', 'LF', 'CF', 'RF',  'RW','LAM', 'CAM', 'RAM', 'LM', 'LCM', 'CM', 'RCM', 'RM', 'LWB', 'LDM','CDM', 'RDM', 'RWB', 'LB', 'LCB','CB', 'RCB', 'RB'], axis=1,inplace=True)

# 查看剩下的特征名称
datafr.columns
```

![](17.jpg)

然后需要对剩下的特征做一些基本处理，比如离散化编码，处理缺失值等。

```python
attributes = datafr.iloc[:, 14:]
attributes['Skill Moves'] = datafr['Skill Moves']
attributes['Age'] = datafr['Age']
# 独热编码
workrate = datafr['Work Rate'].str.get_dummies(sep='/ ')
attributes = pd.concat([attributes, workrate], axis=1)
df = attributes
# 删掉有缺失值的数据
attributes = attributes.dropna()
df['Name'] = datafr['Name']
df['Position'] = datafr['Position']
df = df.dropna()

print(attributes.columns)
```

![](18.jpg)

接着看一下这些特征的相关性。

```python
plt.figure(figsize=(9,9))

# 计算特征之间的协方差
corr = attributes.corr()

# 因为相关性的热力图是对角线对称的，所以把上三角形遮掉看起来会清晰点
mask = np.zeros_like(corr, dtype=np.bool)
mask[np.triu_indices_from(mask)] = True

# 画相关性热力图
sns.heatmap(corr, mask=mask, cmap="RdBu", vmax=.3, center=0,
            square=True, linewidths=.7, cbar_kws={"shrink": .7})
```

![](19.jpg)

从上面的相关图可以看出，守门员的许多属性与前锋、中场和后卫的属性呈负相关。而且大部分特征之间的相关性比较小，算是比较独立的特征。

验证完相关性后，需要对特征进行标准化，来减少特征值的量纲所带来的影响，因为k近邻算法对量纲非常敏感。

```python
# 导入sklearn提供的z-score标准化接口
from sklearn.preprocessing import StandardScaler

scaled = StandardScaler()
X = scaled.fit_transform(attributes)
```

最后，可以使用 sklearn 提供的k近邻算法接口来建模。

```python
from sklearn.neighbors import NearestNeighbors

# 使用k近邻算法建模
recommendations = NearestNeighbors(n_neighbors=5,algorithm='kd_tree')
recommendations.fit(X)
```

有了模型之后，就可以实现相似球员的推荐了。

```python
def get_index(x):
    return df[df['Name']==x].index.tolist()[0]

def recommend_similar(player):
    print("These are 4 players similar to {} : ".format(player))
    index=  get_index(player)
    for i in player_index[index][1:]:
        print("Name: {0}\nPosition: {1}\n".format(df.iloc[i]['Name'],df.iloc[i]['Position']))

# 查找与哈扎德相似的4名球员
recommend_similar('E. Hazard')
```

![](20.jpg)

比利时的哈扎德，现在效力于皇家马德里。他的球风和能力与巴西的国脚内马尔是比较相近的。从结果可以看出，我们的推荐模型还是比较准确的。

